:- module( options, [ 
			  en_list/2,
			  options/2, 
			  options/3,
			  options_append/3, 
			  options_append/4,
			  options_call/2,
			  options_debug/3, 
			  options_propagate/4,
			  options_restore/2,
			  options_return/2,
			  options_version/2
			] ).

/**  <module> Options handling and processing.

This is a stoics.infrastructure pack for handling option arguments.
The main concept is to treat options as naive Prolog lists which the
programmer can manipulate and specialise if they need to, while providing a
small number of predicates that manage basic common operations on options.
Options are particularly important in the context of SWI packs, as making
code publicly available to others often involves allowing for variations
in the behaviour of the code.

The library provides simple extensions to the basic list manipulation
predicates. In many cases it is just the error handling that is the main
difference to standard predicates.

Technically the library is designed on the semantics of memberchk/2. 
Looking for an Option in a list of options, memberchk/2 will return the
leftmost match. Library(options) sees options as a concatenation (append/3)
of the user provided options (arguments for hereon) and the defaults
provided by the predicate.

The default option values for a predicate are given by a predicate
of the same name but postfixed by '_defaults'. The library also allows for
reading user specific default options by reading profiles from a file
located at $HOME/.pl/<pred_name>.pl, if that file exists. Each options file
should have a number of option terms given as facts.

Some distinctive features of pack(options)

  * minimal "magic" behind the scenes

  * allows sloppy (un-listed) single option argument

  * defaults might depend on input

  * can-do cascading  (options of distinct predicates should be disjoint)

  * processing debug(Dbg) terms which optionise calls to debug/1

  * uniform access to user specific file-based default options

For an example see program options_example_sort_defaults.pl in examples directory.

==
?- edit( pack(options/examples/ex_sort) ).

?- [pack(options/examples/ex_sort)].

?- ex_sort( [a,b,e,c,b], Ord, true ).
Ord = [a, b, c, e].
?- ex_sort( [a,b,e,c,b], Ord, debug(true) ).
% Input list length: 5
% Output list length: 4
Ord = [a, b, c, e].
?- ex_sort( [a,b,e,c,b], Ord, order(>) ).
Ord = [e, c, b, a].
?- ex_sort( [a,b,e,c,b], Ord, duplicates(true) ).
Ord = [a, b, b, c, e].
==

Create file $HOME/.pl/ex_sort.pl with content
order(>). 

==
?- ex_sort( [a,b,e,c,b], Ord, true ).
Ord = [e, c, b, a].
==
Default for user is now order(>) which can still be over-ridden at invocation
==
?- ex_sort( [a,b,e,c,b], Ord, order(<) ).
Ord = [a, b, c, e].
==

Predicates

  * en_list/2,
  * options/2,3
  * options_append/3,4
  * options_propagate/4
  * options_restore/2
  * options_return/2

@author   nicos angelopoulos
@version  0.2.0 2015/7/5
@version  0.4.0 2016/2/29
@tbd      type checking (see options_propagate for a first use)

*/

:- use_module(library(requires)).
:- lib(pack_errors).

% :- requires( options:options/2 ).

/** options_version( -Version, -Date ).

Current version and release date for the library.

*/
options_version( 0:4:0, date(2016,2,29) ).


options_defaults( [rem_opts(_),en_list(false),ground(false)] ).

%% options( +Required, +Opts ).
%% options( +Required, +Opts, +OptionsOpts ).
%
% This should more naturally be option/2 but as this might cause confusion with
% Swi's own option library.
% Required can be a single or list of terms. 
%
% Opts can also be non list as it is passed through en_list/2.
% 
% OptionsOpt
% * en_list(Enlist=false)
%    when true it enlists the returning argument
% 
% * ground(Ground=false)
%    should the selected option required to be ground, and what to do if not
%    * false
%      dont do anything
%    * true
%      fail
%    * error
%      throw error
% 
% * true(TruthValue)
%   only variable is allowed, if present the call always succeeds and TruthValue is bound to either _true_ or _false_.
%   
% * rem_opts(RemOpts)
%    RemOpts is Opts after all functor/shape matching mentions of Required have been removed.
%
% 
% The predicate fails silently if the first Required option with equivalent "shape"
% in Opts fails to pattern match (unlike classic memberchk/2).
%
%==
% ?- options( x(y), [x(x)] ).
% false.
% 
% ?- options( x(y), [x(x),x(y)] ).
% false.
%
% ?- options( x(X), [a(b),c(d),x(y),x(x)] ).
% X=y.
%
% ?- options( [a(X),c(b)], [a(b),x(x),b(c)] ).
% ERROR: Required option: c(b), not present in options: [a(b),x(x)]
% % Execution Aborted
% 
% ?- options( x(X), [a(b),c(d),x(y),x(x)], rem_opts(Rem) ).
% X = y,
% Rem = [a(b), c(d)].
%
% ?- options( x(X), [a(b),c(d),x(y),x(x)], en_list(true) ).
% X= [y].
%
% ?- options( a(A), [a(X),b(c)], ground(error) ).
% ERROR: pack(options): Option should be ground, but found: a(_G1470), in options: [a(_G1470),b(c)]
% ?- options( a(A), [a(X),b(c)], ground(true) ).
% false.
% 
% ?- options( a(A), [a(X),b(c)] ).
% A = X.
% 
% ?- options( a(A), [a(X),b(c)], ground(false) ).
% A = X.
% 
% ?- options( a(b), [a(a),a(b),b(c)], true(T) ).
% T = fail.
% 
% ?- options( a(A), [a(a),a(b),b(c)], true(T) ).
% A = a,
% T = true.
% 
% ?- options( a(a), [a(a),a(b),b(c)], true(T) ).
% T = true.
% 
%==
%
% @author nicos angelopoulos
% @version  0.2 2015/01/16
% @version  0.3 2015/12/06  changed 3rd argument to Options of its own invocation
%
options( Needed, Opts ) :-
	options( Needed, Opts, [] ).

options( ReqS, OptS, OwnOptS ) :-
	options_append( options, OwnOptS, OwnOpts ),
	en_list( ReqS, Reqs ),
	en_list( OptS, Opts ),
	memberchk( en_list(Enlist), OwnOpts ),
	debug( fixme, 'Use type checking on: ~a', 'Enlist' ),
	options_required_truth_value( Reqs, Opts, Enlist, OwnOpts, Rem ),
	memberchk( rem_opts(Rem), OwnOpts ).

options_required_truth_value( Reqs, Opts, Enlist, OwnOpts, Rem ) :-
	options_required( Reqs, Opts, Enlist, OwnOpts, Rem ),
	!,
	( memberchk(true(true),OwnOpts) -> true; true ).
options_required_truth_value( _Reqs, Opts, _Enlist, OwnOpts, Opts ) :-  % Rem = Opts here
	memberchk( true(fail), OwnOpts ).

options_required( [], Opts, _Enlist, _OwnOpts, Opts ).
options_required( [Needed|T], Opts, Enlist, OwnOpts, Rem ) :-
	options_option_required( Opts, Needed, Enlist, OwnOpts, Rest ),
	options_required( T, Rest, Enlist, OwnOpts, Rem ).

options_option_required( Opts, Needed, Enlist, OwnOpts, Rest ) :-
	term_template( Needed, Termlate ), % we can use select_first/3 here
	select_first( Opts, Termlate, Rest ),
	!,
	memberchk( ground(Ground), OwnOpts ),
	holds( ground(Termlate), IsGround ),
	option_arg_ground( IsGround, Ground, Termlate, Opts ),
	option_arg_enlist( Enlist, Termlate, Needed ).
options_option_required( Opts, Needed, _Enlist, _OwnOpts, _Rest ) :-
	/*
	Mess = 'Required option: ~w, not present in options: ~w',
	debug( fixme, 'Use pack(pack_debug).', _ ),
	message_report( Mess, [Needed,Opts], error ),
	abort.
	*/
	throw( pack_error(options,opt_required(Needed,Opts)) ).

/** option_arg_ground( +Ground, ?Termlate ).

Check groundness of found Option.

*/
option_arg_ground( true, _Ground, _Termlate, _Opts ).
option_arg_ground( false, Ground, Termlate, Opts ) :-
	options_arg_non_ground( Ground, Termlate, Opts ).

% options_arg_non_ground( true, Termlate, Opts ) :- fail
options_arg_non_ground( false, _Termlate, _Opts ).  % Termlate is NOT required to be ground
options_arg_non_ground( error, Needed, Opts ) :-
	throw( pack_error(options,opt_mustbe_ground(Needed,Opts)) ).

option_arg_enlist( true, Termlate, Needed ) :-
	Termlate =.. [Name,Arg1|Args],
	en_list( Arg1, Enlist1 ),
	Needed =.. [Name,Enlist1|Args].
option_arg_enlist( false, Needed, Needed ).

%% en_list( +Term, -Listed ).
%
% Ensure that Term is either a list of things or 
% a non-var term that that is wrapped to a singleton list.
% If In is a variable a ball is thrown.
%
%==
% ?- en_list( x(y), Opts ).
% Opts = [x(y)].
%
% ?- en_list( [x,y], Opts ).
% Opts = [x, y].
%
% ?- en_list( X, Opts ).
% ERROR: Unhandled exception: en_list(encoutered_variable_in_1st_arg(_))
%==
%
en_list( In, Opts ) :-
	is_list( In ),
	!,
	Opts = In.
en_list( In, Opts ) :-
	\+ var( In ),
	!,
	Opts = [In].
en_list( Else, _Opts ) :-
	throw( en_list(2,encountered_variable_in_1st_arg(Else)) ).


% :- requires( debugging_status/2 ).


options_append_known_process_option( debug ).

/** options_append( +PredName, +OptS, -All ).
    options_append( +PredName, +OptS, -All, +OAopts ).

Look for PredName_defaults/1 and if that exists append it to OptS to get All.
OptS is casted to a list before the append, so single terms are allowed as options.
Listens to debug(options_append).
 
The predicate can process debug(Dbg) a commonly used option. Default should be provided by PredName caller.
The infrastructure allows for other options to be added easily.
 
OAopts term or list of: 

  * extra_arg(Arg)          
    multiple allowed. All Arg(s) so passed are added to the Args passed to the defaults predicate, 
    but not to the generated options. Allows for instance to pass argments of the call itself 
    to the defaults predicate without those arguments being added to the Options list

  * foreign(Foreign)
    instantiates to all options that do not have matching default term structure

  * debug(Dbg=none)   
    if _true_ debug this call, and call prolog_debug:debug_topic(Pname). 
    if _false_ trun debugging off for this call.
    Else Dbg can be a debug term (other than none,false,true) or list of debug terms.
 
  * funnel(Proccess)    
    as process() below, but leaves processed options in All.
  * process(Proccess)
    with Proccess in 
    * debug     will turn on debugging according to debug/0,1,2 options, see below

 When processing debugging options in All, the first matching term of the following is used: 
    * debug
      short for debug(true)
   
    * debug(Dbgs)
      short for debug(Dbgs,_Prior)

    * debug(Dbgs,Prior)
      Prior is the prior status of debug(PredName). For each element of Dbgs call RHS:
    * true   
      debug(PredName)
    * false  
       nodebug(PredName)
    * none   
      true
    * all    
      debug(_)
    * Other  
      debug(Other)

== 
?- assert( demo_defaults(ls(true)) ).
?- options_append( demo, ls(false), All ).
All = [ls(false), ls(true)].
 
?- options_append( demo, debug, All, process(debug) ).
All = [ls(true)].
 
?- options_append( demo, [debug(true),ls(false)], All, [process(debug),debug(true)] ).
Turning debugging on for predicate handle: demo
All = [ls(false), ls(true)].
 
% Note that the debug(options_append) has been removed.
?- options_append( demo, debug, All, process(debug) ).
All = [ls(true)].
==
  
The default OAopts list is [funnel(debug)].
  
@author nicos angelopoulos
@version  0.2 2014/9/20
@tbd  add option! for making sure that only recognised options (with additional,"silent" defaults) are accepted ??
@see ~/bin/cline/keep.pl for debug option example
@tbd allow for strict lists inputs
*/
options_append( Pname, Args, Opts ) :-
	options_append( Pname, Args, Opts, [funnel(debug)] ).
options_append( Pname, ArgS, Opts, OAoptS ) :-
	atom_concat( Pname, '_defaults', Dname ),
	en_list( ArgS, Args ),
	options_append_profile_options( Pname, Args, ProfArgs ),
	en_list( OAoptS, OAopts ),
	% options_append_args( OAopts, Args, Arity ),
	% options_def_append( Dname, Pname, Opts, Args, Arity, Defs, All ),
	findall( Xarg, member(extra_arg(Xarg),OAopts), Xargs ),
	append( ProfArgs, Xargs, ExtArgs ),
	options_def_append( Dname, Pname, ProfArgs, ExtArgs, Defs, OptsUnpro ),
	options_append_select_own_debug( OAopts, ProcessOpts, Restore ),
	options_append_process( ProcessOpts, OptsUnpro, Defs, Pname, Opts ),
	options_append_restore_debug_status( Restore ).

% options_append_profile_options( Pname, Args, ProfArgs ),
% If file $HOME/.pl/Pname.pl exists, append its terms to Args.
%
options_append_profile_options( Pname, Args, Semi ) :-
	atomic_list_concat( ['$HOME/.pl/',Pname,'.pl'], Hfile ),
	expand_file_name( Hfile, [File] ),
	exists_file( File ),
	read_file_to_terms( File, Terms, [] ),
	append( Args, Terms, Semi ),
	!.
options_append_profile_options( _Pname, Args, Args ).

/*
options_append_args( Opts, ArgsList, Arity ) :-
	memberchk( args(Args), Opts ),
	!,
	ArgsList = [Args],
	Arity is 2.
options_append_args( _Opts, Args, Arity ) :-
	Args = [],
	Arity is 1.
	*/

options_append_process( [], Opts, _Defs, _Pname, Opts ).
options_append_process( [extra_arg(_)|T], All, Defs, Pname, Opts ) :-
	!,
	options_append_process( T, All, Defs, Pname, Opts ).
options_append_process( [process(Opt)|T], All, Defs, Pname, Opts ) :-
	!,
	select_all( T, process(Opt), _, Rem ), % remove dublicates % should Opt by anonymous here?
	options_append_process_option( Opt, All, Pname, Nxt, _Enh, Opts ),
	options_append_process( Rem, Nxt, Defs, Pname, Opts ).
options_append_process( [funnel(Opt)|T], All, Defs, Pname, Opts ) :-
	!,
	select_all( T, process(Opt), _, Rem ),
	options_append_process_option( Opt, All, Pname, _Nxt, Enh, Rem ),
	options_append_process( Rem, Enh, Defs, Pname, Opts ).
options_append_process( [foreign(Fgn)|T], All, Defs, Pname, Opts ) :-
	!,
	select_all( T, foreign(_), _, Rem ), % remove dublicates 
	exclude( template_in_defaults(Defs), All, Fgn ),
	options_append_process( Rem, All, Defs, Pname, Opts ).
options_append_process( [Opt|_T], _All, _Defs, Pname, _Opts ) :-
	throw( unknown_option_in_options_append(Opt,Pname) ). % fixme

template_in_defaults( Defs, Term ) :-
	term_template( Term, Template ),
	memberchk( Template, Defs ).

options_append_process_option( Opt, All, Pname, Nxt, Enh, Opts ) :- 
	options_append_known_process_option( Opt ),
	!,
	options_append_option_process( Opt, All, Pname, Nxt, Enh, Opts ).
options_append_process_option( Opt, _All, _Pname, _Nxt, _Enh, _Opts ) :- 
	throw( options_append( unknown_process_option(Opt)) ).

% user + program can use debug/0,1,2 the first one 
% is only used, the Dbg terms argument can be a list if 
% more subjects need debugging
% 
% fixme: this doesn't work properly from multi_debugs
%  
options_append_option_process( debug, All, Pname, NxtEnh, Enh, _Opts ) :-
	partition( name_term(debug), All, Dbgs, Nxt ),
	Dbgs = [Dbg|_],
	!,
	( debugging(Pname) -> Status = true; Status = false ),
	Rst = '$restore'(Pname,debug,Status),
	Enh = [Rst|All],
	NxtEnh = [Rst|Nxt],
	options_append_option_process_debug( Dbg, Pname ).
% next clause states that we shouldn't complain if there is no debug/1,2,3
options_append_option_process( debug, All, _Pname, Nxt, Enh, _Opts ) :-
	% fixme: add option_append option for strictness here ?
	Nxt = All,
	Enh = All.
	
options_append_option_process_debug( debug, Pname ) :-
	options_append_option_process_debug_arg( Pname, true ).
options_append_option_process_debug( debug(DbgS), Pname ) :-
	en_list( DbgS, Dbgs ),
	maplist( options_append_option_process_debug_arg(Pname), Dbgs ).
options_append_option_process_debug( debug(DbgS,Prior), Pname ) :-  % fixme, this is now dealt by '$restore' ?
	en_list( DbgS, Dbgs ),
	debugging_status( Pname, Prior ),
	maplist( options_append_option_process_debug_arg(Pname), Dbgs ).

options_append_option_process_debug_arg( _, Var ) :-
	var( Var ),
	!,
	throw( options_append(instantiation_of_debug) ).

/*
options_append_option_process_debug_arg( Pname, _ ) :- !,
	\+ prolog_debug:debug_topic(Pname).
	*/
options_append_option_process_debug_arg( _Pname, none ) :- !.
options_append_option_process_debug_arg( Pname, false ) :- !,
	nodebug( Pname ).
options_append_option_process_debug_arg( Pname, true ) :- !,
	( prolog_debug:debug_topic(Pname) -> true; prolog_debug:debug_topic(Pname) ),
	debug( options_append, 'Turning debugging on for predicate handle: ~w', [Pname] ),
	debug( Pname ).
options_append_option_process_debug_arg( _Pname, all ) :- !,
	debug( options_append, 'Turning indescriminate debugging on', [] ),
	debug( _ ).
options_append_option_process_debug_arg( _, Other ) :- !,
	debug( options_append, 'Turning debugging on for handle: ~w', [Other] ),
	debug( Other ).

options_def_append( Dname, Pname, Args, ExtArgs, Defs, All ) :-
	predicate_name_defined_in( Pname, Mod ),
	debug( options, 'Defined in: ~w', [Mod:Pname] ),
	member( Arity-Gargs, [1-[DefS],2-[ExtArgs,DefS]] ),
	current_predicate( Mod:Dname/Arity ),
	!,
	% append( Left, [DefS], Args ),
	Goal =.. [Dname|Gargs],
	call( Mod:Goal ),
	en_list( DefS, Defs ),
	append( Args, Defs, All ).
options_def_append( Dname, _Pname, Args, Defs, All ) :-
	current_predicate( Dname/Arity ),
	member( Arity-Gargs, [1-[DefS],2-[Args,DefS]] ),
	!,
	% append( Left, [DefS], Args ),
	Goal =.. [Dname|Gargs],
	call( Goal ),
	en_list( DefS, Defs ),
	append( Args, Defs, All ).
options_def_append( Dname, Pname, Opts, [], All ) :-
	Mess = 'Could not locate defaults predicate: ~w, in modules: ~w',
	message_report( Mess, [Dname/'1,2',[user,Pname]], warning ),
	All = Opts.

options_append_select_own_debug( Opts, NoDebug, Restore ) :-
	select_first( Opts, debug(Dbg), NoDebug ),
	memberchk( Dbg, [true,false] ),
	!,
	debugging_status( options_append, Restore ),
	options_append_debug_own( Dbg ).
options_append_select_own_debug( Opts, Opts, true ).

options_append_debug_own( true ) :-
	debug( options_append ).
options_append_debug_own( false ).

options_append_restore_debug_status( true ).
options_append_restore_debug_status( false ) :-
	nodebug( options_append ).

predicate_name_defined_in( Pname, Mod ) :-
	current_module( Mod ),
	current_predicate( Mod:Pname/_).
predicate_name_defined_in( Pname, Mod ) :-
	current_module(In),
	current_predicate(In:Pname/N),
	length(List,N),
	Goal =.. [Pname|List],
	predicate_property(In:Goal,imported_from(Mod)).

%% options_debug( +Format, +Args, +Opts ).
%
% Call debug(_,Format,Args) iff debug(true) is the first debug(_) term in 
% list Opts
%==
% ?- options_debug( 'A simple message at: ~w', noon, [debug(true)] ).
% ?- options_debug( 'A simple message at: ~w', noon, [] ).
%==
%
% @author nicos angelopoulos
% @version  0.1 2015/1/15
%
options_debug( Format, Args, Opts ) :-
	memberchk( debug(Dbg), Opts ),
	!,
	ground( Dbg ), % fixme: report rather than just failing...
	option_boolean_debug( Dbg, Format, Args ).
options_debug( _Format, _Args, _Opts ).

% should we allow debug reporting for self here ?
option_boolean_debug( true, Format, Args ) :-
	!,
	debug( _, Format, Args ).
option_boolean_debug( _, _Format, _Args ).


:- use_module( library(options) ).

options_propagate_defaults( match(first) ).

/** options_propagate( +Onames, +OptsIn, -OptsOut, +PropOpts ).

Propagate all options with names in Onames from OptsIn to OptsOut.
Onames can be a sinle un_listed option name (en_list/2).

The predicate does not check arities.

The order in OptsOut follows the order in Onames.

PropOpts control the predicate behaviour
  * match(Match=first)
     duplicates are removed, use match(all) for no removal

@author nicos angelopoulos
@version  0.1 2015/3/19
@tbd type checking of Match via options_append/4.

*/

options_propagate( OnameS, Ins, Outs, Args ) :-
	en_list( OnameS, Onames ),
	options_append( options_propagate, Args, Opts ),
	options( match(Match), Opts ),
	options_propagate_list( Onames, Match, Ins, Outs ).

options_propagate_list( [], _Match, _Ins, [] ).
options_propagate_list( [H|T], Match, Ins, Outs ) :-
	options_propagate_match( Match, H, Ins, ToOuts ),
	append( ToOuts, TailOuts, Outs ),
	options_propagate_list( T, Match, Ins, TailOuts ).

options_propagate_match( first, H, Ins, ToOuts ) :-
	( (member(Opt,Ins),functor(Opt,H,_)) ->
		ToOuts = [Opt]
		;
		ToOuts = []
	).
options_propagate_match( all, H, Ins, ToOuts ) :-
	findall( Opt, (member(Opt,Ins),functor(Opt,H,_)), ToOuts ).

/** options_restore( +Self, +Opts ).

Restore all options_append/3, auto processed options.

This predicate is often called at the end of a deterministic 
call that depends on a set of options.

Currently the only option that is restored is
 
  * debug(Dbg)

Internals;
  The predicate looks for any macthing '$restore'(Self,OptName,Status), 
  so it should be extensible to other state-based options processing.

*/

options_restore( Self, Opts ) :-
	maplist( option_restore(Self), Opts ).

option_restore( Self, '$restore'(Self,Process,Status) ) :-
	!,
	option_process_restore( Process, Status, Self ).
option_restore( _Self, _Opt ).

option_process_restore( debug, Status, Self ) :-
	option_process_restore_debug( Status, Self ).

option_process_restore_debug( true, Self ) :-
	debug( Self ).
option_process_restore_debug( false, Self ) :-
	nodebug( Self ).

%% options_return( +TermS, +OptS ).
%
% The predicate always succeeds. But if Opts memberchk/2 with Term
% then it would instantiate TermS (a non-list term or list of terms).
% The predicate, thus the name, can be used to return optional information
% to the caller.
%
% OptS and TermS are passed through en_list/2.
%
%==
% ?- options_return( cnm(Cnm), [abc(x),cnm(symbols)] ).
% Cnm = symbols.
% ?- options_return( [data(Dt),cnm(Cnm)], [abc(x),cnm(symbols)] ).
% Cnm = symbols
%==
%
options_return( TermS, OptS ) :-
	en_list( OptS, Opts ),
	en_list( TermS, Terms ),
	options_return_lists( Terms, Opts ).

options_return_lists( [], _Opts ).
options_return_lists( [Term|T], Opts ) :-
	( memberchk( Term, Opts ) -> true; true ),
	options_return_lists( T, Opts ).



options_call_defaults( call_module(user) ).

/** options_call( +Goal, +Opts ).

Call Goal by extending it with a first argument that comes from
selecting a homonymous option from Opts.

Opts 
 * call_module(Mod=user)  
   the module in which to call extended Goal
 
==
?- options_call( plus(2,X), plus(1) ).
X = 3
==

@author nicos angelopoulos
@version  0.1 2015/12/15
@ allow for , separate goals ? 

*/
options_call( Goal, Args ) :-
	options_append( options_call, Args, Opts ),
	compound( Goal, Pname, Grgs ),
	compound(  Opt, Pname, [Org] ),
	options( Opt, Opts ),
	% it is safe to use =.. now as we have at 1 arg
	Callable =.. [Pname,Org|Grgs],
	options( call_module(Mod), Opts ),
	call( Mod:Callable ).

:- multifile( pack_errors:message/3 ).

pack_errors:message( opt_mustbe_ground(Opt,Opts) ) -->
	['Option should be ground, but found: ~w, in options: ~w'-[Opt,Opts]].

pack_errors:message( opt_required(Opt,Opts) ) -->
	['Required option: ~w, not present in options: ~w' - [Opt,Opts]].

pack_errors:message( opt_mismatch(Opt,Vals,Opts) ) -->
	Term =.. [Opt,Val],
	( memberchk(Term,Opts) -> true; Val = '$internal_error_at_opt_mismatch' ),
	['Option: ~w, expects values in: ~w, but found: ~w in options: ~w' 
	 - [Opt,Vals,Val,Opts]].

%% term_template( +Term, -Arity, -Template ).
%% term_template( +Term, -Template ).
%
% Template has the same Arity and functor as Term, but all 
% its arguments are fresh variables.
%
%==
% term_template( t(a,b,c), Arity, Template ).
% Arity = 3,
% Template = t(_G6305, _G6306, _G6307).
%==
%
% @author nicos angelopoulos
% @version  0.1 this used to be template/2
% @version  0.2 2014/4/15, renamed and added /3 version.
%
term_template( Term, Template ) :-
	term_template( Term, _Arity, Template ).

term_template( Term, Arity, Template ) :-
     functor( Term, Name, Arity ),
     functor( Template, Name, Arity ).


%% select_first( +List, +Elem, -Rem ).
%
% An idiom of select_all/4 which unfolds to select_all( List, Elem, [H|_], Rem ), H = Elem.
%==
% ?- select_first( [dbg(t),dbg(f),etc(x)], dbg(W), Rem ).
% W = t,
% Rem = [etc(x)].
%==
% @author nicos angelopoulos
% @version  0.1 2014/4/7
%
select_first( List, Elem, Rem ) :-
	select_all( List, Elem, [H|_], Rem ),
	H = Elem.

/** holds( +Goal, -Holds ).
    
    Goal is called deterministacally with =|Holds = true|= iff Goal
    succeeds. Else, =|Holds = false|=.

    Note that if Holds is instantiated, Goal will still be called.

==
?- holds( X=3, Holds ).
X = 3,
Holds = true.

?- holds( 4=3, Holds ).
Holds = false.

?- holds( member(X,[a,b]), Holds ).
X = a,
Holds = true.


?- holds( member(X,[a,b]), non_true ).
false.

?- holds( (write(x),nl), non_true ).
x
false.
==

@author nicos angelopoulos
@version  0.1 2015/12/9

*/
holds( Goal, Holds ) :-
	call( Goal ),
	!,
	Holds = true.
holds( _Goal, false ).

%% message_report( Format, Args, Kind ).
%
%  An Swi shortcut for printing messages.
%  The predicate first phrases onto a list the Format message
%  filled by Args, as it would do for debug( _, Format, Args ), 
%  then prints these lines as of Kind (error,warning,debug(_)).
%
%==
% ?- Mess = 'Destination:~w already pointed to:~w, repointing to:~w',
% |    F1 = 'file1', F2 = file2, F3 = file3,
% |    message_report( Mess, [F1,F2,F3], warning ).
%
% Warning: Destination:file1 already pointed to:file2, repointing to:file3
% 
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/02/28
%
message_report( Format, Args, Kind ) :-
	phrase('$messages':translate_message(debug(Format,Args)), Lines),
	print_message_lines(current_output, kind(Kind), Lines).

%% select_all( +List, +Elem, -Select, -Rem ).
% 
% Select all elements of List that are term subsumed (subsumes_term/2) by Elem.
% Rem is the non selected elements of List
% 
% works on Swi have n't tested Yap...
% 
% ==
% select_all( [a(b),b(c),a(b),d(a),a(c)], a(A), Sel, Rem ).
% Sel = [a(b), a(b), a(c)],
% Rem = [b(c), d(a)].
% 
% select_all( [a(b),b(c),a(b),d(a),a(c)], a(b), Sel, Rem ).
% Sel = [a(b), a(b)],
% Rem = [b(c), d(a), a(c)].
% 
% ==
% 
% @author nicos angelopoulos
% @version  0.2 2014/4/7
%
% :- ensure_loaded( library(terms) ).

select_all( [], _Elem, [], [] ).
select_all( [H|T], Elem, Sel, NonSel ) :-
     subsumes_term( Elem, H ),
     !,
     Sel = [H|TSel],
     select_all( T, Elem, TSel, NonSel ).
select_all( [H|T], Elem, Sel, [H|TNonSel] ) :-
     select_all( T, Elem, Sel, TNonSel ).



%% name_term( Name, Term ).
%
% True iff Name is the name of the Term.
% Suitable for meta-calls on list of terms. 
%
%== 
% partition( name_term(dev), [dev(x),abc(d)], Yes, No ).
% Yes = [dev(x)],
% No = [abc(d)].
% 
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/03/11
% @see term_name/2. It is probably better and argument order is correct.
%
name_term( Name, Term ) :-
	compound( Term, Name, _Arity ),
	!.
name_term( Name, Term ) :-
	atomic( Term ),
	Name = Term.

%% compound( +Term, -Name, -Args ).
%
%  Tries to deal with syntax changes that allow a() as a legal term.
%
% Examples run on Swi.7 
%==
%  compound( abc, Name, Args ).
%  false.
%  
%  compound( abc(a,b,c), Name, Args ).
% Name = abc,
% Args = [a, b, c].
% 
% compound( Term, abc, [a,b,c] ).
% Term = abc(a, b, c).
% 
% compound( Term, abc, [] ).
% Term = abc().
%
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/1/10 (round about)
%
compound( Term, Name, Args ) :-
	current_predicate( compound_name_arguments/3 ),
	!,
	once( (compound(Term) ; (ground(Name),is_list(Args))) ),
	% !,
	compound_name_arguments( Term, Name, Args ).
compound( Term, Name, Args ) :-
	once( (compound(Term) ; (ground(Name),ground(Args))) ),
	Term =.. [Name,Args].

